@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2015	PgPool Global Development Group
+ * Copyright (c) 2003-2017	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -132,7 +132,8 @@ static char* process_name_from_pid(pid_t pid);
 
 static struct sockaddr_un un_addr;		/* unix domain socket path */
 static struct sockaddr_un pcp_un_addr;  /* unix domain socket path for PCP */
-ProcessInfo *process_info = NULL;	/* Per child info table on shmem */
+ProcessInfo *process_info = NULL;		/* Per child info table on shmem */
+struct timeval random_start_time;
 
 /*
  * Private copy of backend status
@@ -191,7 +192,10 @@ int PgpoolMain(bool discard_status, bool clear_memcache_oidmaps)
 
 	MemoryContext MainLoopMemoryContext;
 	sigjmp_buf	local_sigjmp_buf;
-	
+
+	/* For PostmasterRandom */
+	gettimeofday(&random_start_time, NULL);
+
 	/* Set the process type variable */
 	processType = PT_MAIN;
 	processState = INITIALIZING;
@@ -763,7 +767,7 @@ static int *create_inet_domain_sockets(const char *hostname, const int port)
 			errdetail("asprintf() failed: %s", strerror(errno))));
 	}
 
-	if ((ret = getaddrinfo(hostname, portstr, &hints, &res)) != 0)
+	if ((ret = getaddrinfo((!hostname || strcmp(hostname, "*") == 0) ? NULL : hostname, portstr, &hints, &res)) != 0)
 	{
 		ereport(FATAL,
 			(errmsg("failed to create INET domain socket"),
@@ -1020,7 +1024,7 @@ static void terminate_all_childrens()
     {
 		int ret_pid;
 
-        wpid = waitpid(-1, &ret_pid, WNOHANG);
+        wpid = waitpid(-1, &ret_pid, 0);
     } while (wpid > 0 || (wpid == -1 && errno == EINTR));
 
     if (wpid == -1 && errno != ECHILD)
@@ -1288,7 +1292,7 @@ static RETSIGTYPE exit_handler(int sig)
     {
 		int ret_pid;
 
-        wpid = waitpid(-1, &ret_pid, WNOHANG);
+        wpid = waitpid(-1, &ret_pid, 0);
     } while (wpid > 0 || (wpid == -1 && errno == EINTR));
 
     if (wpid == -1 && errno != ECHILD)
@@ -1345,9 +1349,14 @@ static RETSIGTYPE failover_handler(int sig)
 
 	POOL_SETMASK(&BlockSig);
 	failover_request = 1;
+
+	write(pipe_fds[1], "\0", 1);
+
+#ifdef NOT_USED
 	if(write(pipe_fds[1], "\0", 1) < 0)
         ereport(WARNING,
                 (errmsg("failover_handler: write to pipe failed with error \"%s\"", strerror(errno))));
+#endif
 
 	POOL_SETMASK(&UnBlockSig);
 
@@ -1370,6 +1379,7 @@ static void failover(void)
 	int status;
 	int sts;
 	bool need_to_restart_pcp = false;
+	bool all_backend_down = true;
 
 	ereport(DEBUG1,
 		(errmsg("failover handler called")));
@@ -1487,6 +1497,20 @@ static void failover(void)
 					 BACKEND_INFO(node_id).backend_hostname,
 					 BACKEND_INFO(node_id).backend_port)));
 
+			/* Check to see if all backends are down */
+			for (i=0;i<NUM_BACKENDS;i++)
+			{
+				if (BACKEND_INFO(i).backend_status != CON_DOWN &&
+					BACKEND_INFO(i).backend_status != CON_UNUSED)
+				{
+					ereport(LOG,
+							(errmsg("Node %d is not down (status: %d)",
+									i, BACKEND_INFO(i).backend_status)));
+					all_backend_down = false;
+					break;
+				}
+			}
+
 			BACKEND_INFO(node_id).backend_status = CON_CONNECT_WAIT;	/* unset down status */
 			(void)write_status_file();
 
@@ -1630,11 +1654,22 @@ static void failover(void)
 		* attached node, but load balanced node is not changed until this
 		* session ends, so it's harmless anyway.
 		*/
-		if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP)	&&
-			reqkind == NODE_UP_REQUEST)
+
+		/*
+		 * On 2015/9/21 Tatsuo Ishii says: this judgment is not sufficient if
+		 * all backends were down. Child process has local status in which all
+		 * backends are down. In this case even if new connection arrives from
+		 * frontend, the child will not accept it because the local status
+		 * shows all backends are down. For this purpose we refer to
+		 * "all_backend_down" variable, which was set before updating backend status.
+		 *
+		 * See bug 248 for more details.
+		 */
+
+		if (STREAM && reqkind == NODE_UP_REQUEST &&	all_backend_down == false)
 		{
 			ereport(LOG,
-					(errmsg("Do not restart children because we are failbacking node id %d host%s port:%d and we are in streaming replication mode", node_id,
+					(errmsg("Do not restart children because we are failbacking node id %d host: %s port: %d and we are in streaming replication mode and not all backends were down", node_id,
 					 BACKEND_INFO(node_id).backend_hostname,
 					 BACKEND_INFO(node_id).backend_port)));
 
@@ -2009,9 +2044,17 @@ static RETSIGTYPE reap_handler(int sig)
 
 	POOL_SETMASK(&BlockSig);
 	sigchld_request = 1;
-	if(write(pipe_fds[1], "\0", 1) < 0)
+
+	if (pipe_fds[1])
+	{
+		write(pipe_fds[1], "\0", 1);
+	}
+
+#ifdef NOT_USED
+	if(pipe_fds[1] && write(pipe_fds[1], "\0", 1) < 0)
         ereport(WARNING,
             (errmsg("reap_handler: write to pipe failed with error \"%s\"", strerror(errno))));
+#endif
 
 	POOL_SETMASK(&UnBlockSig);
 
@@ -2267,12 +2310,14 @@ static RETSIGTYPE wakeup_handler(int sig)
 
 	POOL_SETMASK(&BlockSig);
 	wakeup_request = 1;
+
+	write(pipe_fds[1], "\0", 1);
+#ifdef NOT_USED
 	if(write(pipe_fds[1], "\0", 1) < 0)
         ereport(WARNING,
             (errmsg("wakeup_handler: write to pipe failed with error \"%s\"", strerror(errno))));
+#endif
 	POOL_SETMASK(&UnBlockSig);
-
-	errno = save_errno;
 }
 
 /*
@@ -2285,9 +2330,12 @@ static RETSIGTYPE reload_config_handler(int sig)
 
 	POOL_SETMASK(&BlockSig);
 	reload_config_request = 1;
+	write(pipe_fds[1], "\0", 1);
+#ifdef NOT_USED
 	if(write(pipe_fds[1], "\0", 1) < 0)
         ereport(WARNING,
             (errmsg("reload_config_handler: write to pipe failed with error \"%s\"", strerror(errno))));
+#endif
 
 	POOL_SETMASK(&UnBlockSig);
 
